@@ -1,4 +1,5 @@
 #include "filezilla.h"
+#include "activity_logger_layer.h"
 #include "controlsocket.h"
 #include "directorycache.h"
 #include "engineprivate.h"
@@ -32,12 +33,13 @@
 	#endif
 #endif
 
-CControlSocket::CControlSocket(CFileZillaEnginePrivate & engine)
+CControlSocket::CControlSocket(CFileZillaEnginePrivate & engine, bool use_shm)
 	: event_handler(engine.event_loop_)
 	, engine_(engine)
 	, opLockManager_(engine.opLockManager_)
 	, logger_(engine.GetLogger())
 {
+	InitBufferPool(use_shm);
 }
 
 CControlSocket::~CControlSocket()
@@ -252,16 +254,21 @@ int CControlSocket::DoClose(int nErrorCode)
 std::wstring CControlSocket::ConvertDomainName(std::wstring const& domain)
 {
 #ifdef FZ_WINDOWS
-	int len = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), domain.size() + 1, nullptr, 0);
+	if (domain.size() > 1204 * 1024) {
+		return domain;
+	}
+
+	int const inlen = static_cast<int>(domain.size() + 1);
+	int len = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), inlen, nullptr, 0);
 	if (!len) {
 		log(logmsg::debug_warning, L"Could not convert domain name");
 		return domain;
 	}
 
 	wchar_t* output = new wchar_t[len];
-	int res = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), domain.size() + 1, output, len);
+	int res = IdnToAscii(IDN_ALLOW_UNASSIGNED, domain.c_str(), inlen, output, len);
 	if (!res) {
-		delete [] output;
+		delete[] output;
 		log(logmsg::debug_warning, L"Could not convert domain name");
 		return domain;
 	}
@@ -368,11 +375,11 @@ int CControlSocket::CheckOverwriteFile()
 	data.localFileTime_ = data.download() ? data.writer_factory_.mtime() : data.reader_factory_.mtime();
 
 	if (data.download()) {
-		if (data.localFileSize_ == aio_base::nosize && data.localFileTime_.empty()) {
+		if (data.localFileSize_ == fz::aio_base::nosize && data.localFileTime_.empty()) {
 			return FZ_REPLY_OK;
 		}
 	}
-	
+
 	CDirentry entry;
 	bool dirDidExist;
 	bool matchedCase;
@@ -402,7 +409,7 @@ int CControlSocket::CheckOverwriteFile()
 		}
 	}
 
-	
+
 
 	auto notification = std::make_unique<CFileExistsNotification>();
 
@@ -451,7 +458,11 @@ void SleepOpData::operator()(fz::event_base const&)
 CFileTransferOpData::CFileTransferOpData(wchar_t const* name, CFileTransferCommand const& cmd)
 	: COpData(Command::transfer, name)
 	, flags_(cmd.GetFlags())
-	, reader_factory_(cmd.GetReader()), writer_factory_(cmd.GetWriter()), localName_(reader_factory_ ? reader_factory_.name() : writer_factory_.name()), remoteFile_(cmd.GetRemoteFile()), remotePath_(cmd.GetRemotePath())
+	, reader_factory_(cmd.GetReader())
+	, writer_factory_(cmd.GetWriter())
+	, localName_(reader_factory_ ? reader_factory_.name() : writer_factory_.name())
+	, remoteFile_(cmd.GetRemoteFile())
+	, remotePath_(cmd.GetRemotePath())
 {
 	localFileSize_ = download() ? writer_factory_.size() : reader_factory_.size();
 	localFileTime_ = download() ? writer_factory_.mtime() : reader_factory_.mtime();
@@ -658,7 +669,7 @@ void CControlSocket::InvalidateCurrentWorkingDir(CServerPath const& path)
 	if (path.empty() || currentPath_.empty()) {
 		return;
 	}
-	
+
 	if (path.IsParentOf(currentPath_, false, true)) {
 		if (!operations_.empty()) {
 			m_invalidateCurrentPath = true;
@@ -708,11 +719,6 @@ CRealControlSocket::~CRealControlSocket()
 	ResetSocket();
 }
 
-bool CRealControlSocket::Connected() const
-{
-	return socket_ ? socket_->is_connected() : false;
-}
-
 int CRealControlSocket::Send(unsigned char const* buffer, unsigned int len)
 {
 	if (!active_layer_) {
@@ -737,7 +743,7 @@ int CRealControlSocket::Send(unsigned char const* buffer, unsigned int len)
 		}
 
 		if (written) {
-			SetActive(CFileZillaEngine::send);
+			RecordActivity(activity_logger::send, written);
 		}
 
 		if (static_cast<unsigned int>(written) < len) {
@@ -838,7 +844,7 @@ int CRealControlSocket::OnSend()
 		}
 
 		if (written) {
-			SetActive(CFileZillaEngine::send);
+			RecordActivity(activity_logger::send, written);
 
 			send_buffer_.consume(static_cast<size_t>(written));
 		}
@@ -869,7 +875,8 @@ int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
 
 	ResetSocket();
 	socket_ = std::make_unique<fz::socket>(engine_.GetThreadPool(), nullptr);
-	ratelimit_layer_ = std::make_unique<fz::rate_limited_layer>(this, *socket_, &engine_.GetRateLimiter());
+	activity_logger_layer_ = std::make_unique<activity_logger_layer>(nullptr, *socket_, engine_.activity_logger_);
+	ratelimit_layer_ = std::make_unique<fz::rate_limited_layer>(this, *activity_logger_layer_, &engine_.GetRateLimiter());
 	active_layer_ = ratelimit_layer_.get();
 
 	const int proxy_type = engine_.GetOptions().get_int(OPTION_PROXY_TYPE);
@@ -894,11 +901,13 @@ int CRealControlSocket::DoConnect(std::wstring const& host, unsigned int port)
 		}
 	}
 
+	SetSocketBufferSizes();
+
 	int res = active_layer_->connect(fz::to_native(ConvertDomainName(host)), port);
 
 	if (res) {
 		log(logmsg::error, _("Could not connect to server: %s"), fz::socket_error_description(res));
-		return FZ_REPLY_DISCONNECTED | FZ_REPLY_ERROR; 
+		return FZ_REPLY_DISCONNECTED | FZ_REPLY_ERROR;
 	}
 
 	return FZ_REPLY_WOULDBLOCK;
@@ -919,6 +928,7 @@ void CRealControlSocket::ResetSocket()
 	// Destroy in reverse order
 	proxy_layer_.reset();
 	ratelimit_layer_.reset();
+	activity_logger_layer_.reset();
 	socket_.reset();
 
 	send_buffer_.clear();
@@ -1007,7 +1017,7 @@ bool CControlSocket::SetFileExistsAction(CFileExistsNotification *pFileExistsNot
 		}
 		break;
 	case CFileExistsNotification::resume:
-		if (data.download() && data.localFileSize_ != aio_base::nosize) {
+		if (data.download() && data.localFileSize_ != fz::aio_base::nosize) {
 			data.resume_ = true;
 		}
 		else if (!data.download() && data.remoteFileSize_ >= 0) {
@@ -1129,10 +1139,10 @@ void CControlSocket::operator()(fz::event_base const& ev)
 		&CControlSocket::OnObtainLock);
 }
 
-void CControlSocket::SetActive(CFileZillaEngine::_direction direction)
+void CControlSocket::RecordActivity(activity_logger::_direction direction, uint64_t amount)
 {
 	SetAlive();
-	engine_.SetActive(direction);
+	engine_.activity_logger_.record(direction, amount);
 }
 
 void CControlSocket::SendDirectoryListingNotification(CServerPath const& path, bool failed)
@@ -1162,6 +1172,47 @@ void CControlSocket::Sleep(fz::duration const& delay)
 	Push(std::make_unique<SleepOpData>(*this, delay));
 }
 
+bool CControlSocket::InitBufferPool(bool use_shm)
+{
+	if (!buffer_pool_) {
+		buffer_pool_.emplace(logger_, 8, 0, use_shm);
+	}
+	return *buffer_pool_;
+}
+
+std::unique_ptr<fz::writer_base> CControlSocket::OpenWriter(fz::writer_factory_holder & factory, uint64_t resumeOffset, bool withProgress)
+{
+	if (!factory || !buffer_pool_) {
+		return {};
+	}
+
+	auto file_writer = dynamic_cast<fz::file_writer_factory*>(&*factory);
+	if (file_writer) {
+		std::wstring tmp;
+		CLocalPath local_path(file_writer->name(), &tmp);
+		if (local_path.HasParent()) {
+			fz::native_string last_created;
+			fz::mkdir(fz::to_native(local_path.GetPath()), true, fz::mkdir_permissions::normal, &last_created);
+			if (!last_created.empty()) {
+				// Send out notification
+				auto n = std::make_unique<CLocalDirCreatedNotification>();
+				if (n->dir.SetPath(fz::to_wstring(last_created))) {
+					engine_.AddNotification(std::move(n));
+				}
+			}
+		}
+	}
+
+	fz::writer_base::progress_cb_t status_update;
+	if (withProgress) {
+		status_update = [&s = engine_.transfer_status_](fz::writer_base const*, uint64_t written) {
+			s.SetMadeProgress();
+			s.Update(written);
+		};
+	}
+	return factory->open(*buffer_pool_, resumeOffset, status_update, buffer_pool_->buffer_count());
+}
+
 int64_t CalculateNextChunkSize(int64_t remaining, int64_t lastChunkSize, fz::duration const& lastChunkDuration, int64_t minChunkSize, int64_t multiple, int64_t partCount, int64_t maxPartCount, int64_t maxChunkSize)
 {
 	if (remaining <= 0) {
