@@ -24,7 +24,7 @@ CHttpFileTransferOpData::CHttpFileTransferOpData(CHttpControlSocket & controlSoc
 }
 
 CHttpFileTransferOpData::CHttpFileTransferOpData(CHttpControlSocket & controlSocket, CHttpRequestCommand const& cmd)
-	: CFileTransferOpData(L"CHttpFileTransferOpData", CFileTransferCommand(writer_factory_holder(), CServerPath(), std::wstring(), transfer_flags::download))
+	: CFileTransferOpData(L"CHttpFileTransferOpData", CFileTransferCommand(fz::writer_factory_holder(), CServerPath(), std::wstring(), transfer_flags::download))
 	, CHttpOpData(controlSocket)
 {
 	reader_factory_ = cmd.body_;
@@ -51,7 +51,7 @@ int CHttpFileTransferOpData::Send()
 		}
 
 		if (reader_factory_) {
-			rr_.request_.body_ = reader_factory_.open(0, engine_, nullptr, aio_base::shm_flag_none);
+			rr_.request_.body_ = reader_factory_->open(*controlSocket_.buffer_pool_, 0, fz::aio_base::nosize, controlSocket_.buffer_pool_->buffer_count());
 			if (!rr_.request_.body_) {
 				return FZ_REPLY_CRITICALERROR;
 			}
@@ -60,7 +60,7 @@ int CHttpFileTransferOpData::Send()
 		opState = filetransfer_transfer;
 		if (writer_factory_) {
 			auto s = writer_factory_.size();
-			if (s != aio_base::nosize) {
+			if (s != fz::aio_base::nosize) {
 				localFileSize_ = static_cast<int64_t>(s);
 			}
 
@@ -149,7 +149,7 @@ int CHttpFileTransferOpData::OnHeader()
 	}
 
 	if (writer_factory_) {
-		auto writer = writer_factory_.open(resume_ ? localFileSize_ : 0, engine_, &controlSocket_, aio_base::shm_flag_none);
+		auto writer = controlSocket_.OpenWriter(writer_factory_, resume_ ? localFileSize_ : 0, true);
 		if (!writer) {
 			return FZ_REPLY_CRITICALERROR;
 		}
