@@ -45,9 +45,6 @@ protected:
 	int AddToStream(std::wstring const& cmd);
 	int AddToStream(std::string_view cmd);
 
-#ifndef FZ_WINDOWS
-	int shm_fd_{-1};
-#endif
 	std::unique_ptr<fz::process> process_;
 	std::unique_ptr<CStorjInputThread> input_thread_;
 
