@@ -20,6 +20,7 @@
 #include "remote_recursive_operation.h"
 #include "sizeformatting.h"
 #include "timeformatting.h"
+#include "../commonui/misc.h"
 
 #include <wx/clipbrd.h>
 #include <wx/menu.h>
@@ -85,8 +86,8 @@ public:
 			pDragDropManager->pDropTarget = m_pRemoteListView;
 		}
 
-		auto const format = m_pDataObject->GetReceivedFormat();
-		if (format == m_pFileDataObject->GetFormat() || format == m_pLocalDataObject->GetFormat()) {
+		auto const format = GetReceivedFormat();
+		if (format == m_pFileDataObject->GetFormat() || format == LocalDataObjectFormat()) {
 			std::wstring subdir;
 			int flags = 0;
 			int hit = m_pRemoteListView->HitTest(wxPoint(x, y), flags, 0);
@@ -101,24 +102,24 @@ public:
 					}
 				}
 			}
-
 			if (format == m_pFileDataObject->GetFormat()) {
 				m_pRemoteListView->m_state.UploadDroppedFiles(m_pFileDataObject, subdir, false);
 			}
 			else {
-				m_pRemoteListView->m_state.UploadDroppedFiles(m_pLocalDataObject, subdir, false);
+				m_pRemoteListView->m_state.UploadDroppedFiles(GetLocalDataObject(), subdir, false);
 			}
 			return wxDragCopy;
 		}
 
 		// At this point it's the remote data object.
+		auto * obj = GetRemoteDataObject();
 
-		if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId()) {
+		if (obj->GetProcessId() != (int)wxGetProcessId()) {
 			wxMessageBoxEx(_("Drag&drop between different instances of FileZilla has not been implemented yet."));
 			return wxDragNone;
 		}
 
-		if (m_pRemoteDataObject->GetSite().server != m_pRemoteListView->m_state.GetSite().server) {
+		if (obj->GetSite().server != m_pRemoteListView->m_state.GetSite().server) {
 			wxMessageBoxEx(_("Drag&drop between different servers has not been implemented yet."));
 			return wxDragNone;
 		}
@@ -151,15 +152,15 @@ public:
 		}
 
 		// Make sure target path is valid
-		if (target == m_pRemoteDataObject->GetServerPath()) {
+		if (target == obj->GetServerPath()) {
 			wxMessageBoxEx(_("Source and target of the drop operation are identical"));
 			return wxDragNone;
 		}
 
-		const std::vector<CRemoteDataObject::t_fileInfo>& files = m_pRemoteDataObject->GetFiles();
+		const std::vector<CRemoteDataObject::t_fileInfo>& files = obj->GetFiles();
 		for (auto const& info : files) {
 			if (info.dir) {
-				CServerPath dir = m_pRemoteDataObject->GetServerPath();
+				CServerPath dir = obj->GetServerPath();
 				dir.AddSegment(info.name);
 				if (dir == target) {
 					return wxDragNone;
@@ -173,7 +174,7 @@ public:
 
 		for (auto const& info : files) {
 			m_pRemoteListView->m_state.m_pCommandQueue->ProcessCommand(
-				new CRenameCommand(m_pRemoteDataObject->GetServerPath(), info.name, target, info.name)
+				new CRenameCommand(obj->GetServerPath(), info.name, target, info.name)
 				);
 		}
 
@@ -185,7 +186,7 @@ public:
 
 	virtual bool OnDrop(wxCoord x, wxCoord y)
 	{
-		CScrollableDropTarget<wxListCtrlEx>::OnDrop(x, y);
+		CScrollableDropTarget::OnDrop(x, y);
 		ClearDropHighlight();
 
 		if (!m_pRemoteListView->m_pDirectoryListing) {
@@ -244,7 +245,7 @@ public:
 
 	virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxListCtrlEx>::OnDragOver(x, y, def);
+		def = CScrollableDropTarget::OnDragOver(x, y, def);
 
 		if (def == wxDragError ||
 			def == wxDragNone ||
@@ -280,13 +281,13 @@ public:
 
 	virtual void OnLeave()
 	{
-		CScrollableDropTarget<wxListCtrlEx>::OnLeave();
+		CScrollableDropTarget::OnLeave();
 		ClearDropHighlight();
 	}
 
 	virtual wxDragResult OnEnter(wxCoord x, wxCoord y, wxDragResult def)
 	{
-		def = CScrollableDropTarget<wxListCtrlEx>::OnEnter(x, y, def);
+		def = CScrollableDropTarget::OnEnter(x, y, def);
 		return OnDragOver(x, y, def);
 	}
 
@@ -368,21 +369,18 @@ CRemoteListView::~CRemoteListView()
 int CRemoteListView::OnGetItemImage(long item) const
 {
 	CRemoteListView *pThis = const_cast<CRemoteListView *>(this);
+
 	int index = GetItemIndex(item);
-	if (index < 0) {
+	if (index < 0 || static_cast<size_t>(index) >= pThis->m_fileData.size()) {
 		return -1;
 	}
 
-	if (index >= pThis->m_fileData.size()) {
-		return -1;
-	}
 	int &icon = pThis->m_fileData[index].icon;
-
 	if (icon != -2) {
 		return icon;
 	}
 
-	if (!m_pDirectoryListing || index >= m_pDirectoryListing->size()) {
+	if (!m_pDirectoryListing || static_cast<size_t>(index) >= m_pDirectoryListing->size()) {
 		return -1;
 	}
 
@@ -863,33 +861,6 @@ void CRemoteListView::SetDirectoryListing(std::shared_ptr<CDirectoryListing> con
 	}
 }
 
-// Filenames on VMS systems have a revision suffix, e.g.
-// foo.bar;1
-// foo.bar;2
-// foo.bar;3
-std::wstring StripVMSRevision(std::wstring const& name)
-{
-	size_t pos = name.rfind(';');
-	if (pos == std::wstring::npos || !pos) {
-		return name;
-	}
-
-	if (pos == name.size() - 1) {
-		return name;
-	}
-
-	size_t p = pos;
-	while (++p < name.size()) {
-		wchar_t const& c = name[p];
-		if (c < '0' || c > '9') {
-			return name;
-		}
-	}
-
-	return name.substr(0, pos);
-}
-
-
 void CRemoteListView::OnItemActivated(wxListEvent &event)
 {
 	int const action = options_.get_int(OPTION_DOUBLECLICK_ACTION_DIRECTORY);
@@ -1250,7 +1221,7 @@ void CRemoteListView::OnMenuDownload(wxCommandEvent& event)
 	TransferSelectedFiles(localDir, event.GetId() == XRCID("ID_ADDTOQUEUE"));
 }
 
-void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool queue_only)
+void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool queue_only, transfer_flags custom_flags, transfer_flags custom_flags_mask)
 {
 	bool idle = m_state.IsRemoteIdle();
 
@@ -1307,7 +1278,8 @@ void CRemoteListView::TransferSelectedFiles(const CLocalPath& local_parent, bool
 			}
 			m_pQueue->QueueFile(queue_only, true,
 				name, (name == localFile) ? std::wstring() : localFile,
-				local_parent, m_pDirectoryListing->path, site, entry.size);
+				local_parent, m_pDirectoryListing->path, site, entry.size, CEditHandler::none,
+				QueuePriority::normal, custom_flags, custom_flags_mask);
 			added = true;
 		}
 	}
@@ -1384,7 +1356,8 @@ CServerPath CRemoteListView::MenuMkdir()
 		return CServerPath();
 	}
 
-	m_state.m_pCommandQueue->ProcessCommand(new CMkdirCommand(path));
+	transfer_flags const flags = GetMkdirFlags(m_state.GetSite().server, options_, path);
+	m_state.m_pCommandQueue->ProcessCommand(new CMkdirCommand(path, flags));
 
 	// Return name of the New Directory
 	return path;
@@ -2240,6 +2213,7 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 	pDragDropManager->pDragSource = this;
 	pDragDropManager->site = site;
 	pDragDropManager->remoteParent = m_pDirectoryListing->path;
+	pDragDropManager->dragDataObject = pRemoteDataObject;
 
 	// Add files to remote data object
 	item = -1;
@@ -2278,10 +2252,9 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 
 	CLabelEditBlocker b(*this);
 
-	wxDropSource source(this);
+	DropSource source(this);
 	source.SetData(object);
-
-	int res = source.DoDragDrop();
+	int res = source.DoFileDragDrop();
 
 	pDragDropManager->Release();
 
@@ -2289,9 +2262,13 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 		return;
 	}
 
+#if FZ3_USESHELLEXT || defined(__WXMAC__)
 #if FZ3_USESHELLEXT
 	if (ext) {
 		if (!pRemoteDataObject->DidSendData()) {
+#else
+	if (!source.m_OutDir.IsEmpty()) {
+#endif
 			Site newSite = m_state.GetSite();
 			if (!m_state.IsRemoteIdle() ||
 				!newSite || newSite.server != pRemoteDataObject->GetSite().server ||
@@ -2328,16 +2305,26 @@ void CRemoteListView::OnBeginDrag(wxListEvent&)
 					return;
 				}
 			}
-
+#if FZ3_USESHELLEXT
 			CLocalPath target(ext->GetTarget().ToStdWstring());
+#else
+			// wxMessageBoxEx(source.m_OutDir);
+			CLocalPath target(source.m_OutDir.ToStdWstring());
+#endif
 			if (target.empty()) {
+#if FZ3_USESHELLEXT
 				ext.reset(); // Release extension before the modal message box
 				wxMessageBoxEx(_("Could not determine the target of the Drag&Drop operation.\nEither the shell extension is not installed properly or you didn't drop the files into an Explorer window."));
+#else
+				wxMessageBoxEx(_("Could not determine the target of the Drag&Drop operation."));
+#endif
 				return;
 			}
 
 			TransferSelectedFiles(target, false);
+#if FZ3_USESHELLEXT
 		}
+#endif
 	}
 #endif
 }
@@ -2823,6 +2810,6 @@ void CRemoteListView::OnMenuNewfile(wxCommandEvent&)
 		return;
 	}
 
-	CFileTransferCommand *cmd = new CFileTransferCommand(memory_reader_factory(L"Empty file", std::string_view()), m_pDirectoryListing->path, newFileName, transfer_flags());
+	CFileTransferCommand *cmd = new CFileTransferCommand(fz::view_reader_factory(L"Empty file", std::string_view()), m_pDirectoryListing->path, newFileName, transfer_flags());
 	m_state.m_pCommandQueue->ProcessCommand(cmd);
 }
