@@ -30,10 +30,11 @@ wxStatusBarEx::wxStatusBarEx(wxTopLevelWindow* pParent)
 {
 	m_pParent = pParent;
 	m_columnWidths = 0;
-	Create(pParent, wxID_ANY);
 
 	SetBackgroundStyle(wxBG_STYLE_SYSTEM);
 
+	Create(pParent, wxID_ANY);
+
 #ifdef __WXMSW__
 	m_parentWasMaximized = false;
 
@@ -485,7 +486,8 @@ void CStatusBar::DisplayEncrypted()
 		if (protocol == FTPS || protocol == FTPES || protocol == SFTP || protocol == S3 ||
 				protocol == WEBDAV || protocol == AZURE_BLOB || protocol == AZURE_FILE ||
 				protocol == SWIFT || protocol == GOOGLE_CLOUD || protocol == GOOGLE_DRIVE ||
-				protocol == DROPBOX || protocol == ONEDRIVE || protocol == B2 || protocol == BOX) {
+				protocol == DROPBOX || protocol == ONEDRIVE || protocol == B2 || protocol == BOX ||
+				protocol == RACKSPACE) {
 			encrypted = true;
 		}
 		else if (protocol == FTP && pState->GetSecurityInfo(info)) {
